(*
    Copyright (C) 2012 Christophe Calvès

    This file is part of WC.

    WC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    WC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
*)


open Ast ;;
open Base ;;
open Either ;;

(*************************
 *                       *
 *     Primities         *
 *                       *
 *************************)

let ifcst  = "wcif"
let ifarr  = 3


let primitives = [ (ifcst  , ifarr )
                 ]

(* recusrion *)

let reccst = "rec"

(*************************
 *                       *
 *         CORE          *
 *                       *
 *************************)


type ('a,'b) atom = Var of 'a
                  | Cst of string
                  | Dat of 'b

type ('a,'b) core = Atm of ('a,'b) atom
                  | Fun of 'a * ('a,'b) core
                  | App of ('a,'b) core * ('a,'b) core

let mkVar x   = Atm (Var x)
let mkCst x   = Atm (Cst x)
let mkDat x   = Atm (Dat x)
let mkFun s c = Fun (s,c)
let mkApp c d = App (c,d)

let multiFun l = List.fold_right mkFun l
let multiApp c = List.fold_left  mkApp c


let rec groupSeq l = function
 | []             -> failwith "empty sequence"
 | [ASeq l2]      -> groupSeq l l2
 | [x      ]      -> (List.rev l , x)
 | (ASeq l2) :: q -> groupSeq l (l2 @ q)
 |  x        :: q -> groupSeq (x :: l) q
 

let rec coreOfTerm = function
 | AVar x            -> mkVar x
 | ADat y            -> mkDat y
 | AFun   (l,c)      -> multiFun l (coreOfTerm c)
 | AApp   (c,l)      -> multiApp (coreOfTerm c) (List.map coreOfTerm l)
 | ALetIn (l,c)      -> let (n,e) = List.split l
                        in multiApp (multiFun n (coreOfTerm c)) (List.map coreOfTerm e)
 | ASeq    l         ->    let (l2 , c ) = groupSeq [] l
                        in let params    = List.map (fun x -> ("",x)) l2
                        in coreOfTerm ( ALetIn (params , c) )
 | ALetRecIn (x,s,t) -> let u = AApp ( AVar "rec"
                                     , [ AFun ( [x] , s ) ]
                                     )
                        in coreOfTerm ( ALetIn ( [(x,u)] , t ) )
 | AIf (c,t,e)       -> coreOfTerm (AApp ( AVar ifcst
                                         , [ c
                                           ; AFun ([""] , t)
                                           ; AFun ([""] , e)
                                           ]
                                         )
                                   )

let rec freeVar = function
 | Atm ( Var s ) -> SetString.singleton s
 | Atm ( Cst _ ) -> SetString.empty
 | Atm ( Dat _ ) -> SetString.empty
 | App (c1 , c2) -> SetString.union (freeVar c1) (freeVar c2)
 | Fun (s  , c ) -> SetString.remove s (freeVar c)


let genConstant x =
 let rec aux m = function
   | Atm ( Var s ) -> if SetString.mem s m
                      then Atm ( Var s )
                      else Atm ( Cst s )
   | Fun (s  , c ) -> Fun ( s , aux (SetString.add s m) c)
   | App (c1 , c2) -> App ( aux m c1 , aux m c2 )
   | t             -> t
 in aux SetString.empty x


let eta = function
 | Fun ( s , App ( c , Atm ( Var s' ) ) )
     when (s = s') && (not (SetString.mem s (freeVar c)))
       -> Right c
 | t   -> Left  t


let rec process f t = 
 let t' = ( match t with
              | App ( c , d ) -> App ( process f c , process f d )
              | Fun ( s , c ) -> Fun ( s , process f c )
              | Atm a         -> Atm a
          )
 in ( match f t' with
        | Left  r -> r
        | Right u -> process f u
    )


(*************************
 *                       *
 *        SUPER          *
 *                       *
 *************************)


type ('a,'b) super = SApp of ('a,'b) atom * ('a,'b) super list
                   | SAtm of ('a,'b) atom
let mapSuper f =
 let rec aux = function
  | SApp (a , l) -> SApp (f a, List.map aux l)
  | SAtm  a      -> SAtm (f a)
 in aux


let alphaSuper sl =
    let (_,sigma) = List.fold_left (fun (n,m) s -> (succ n , MapString.add s n m))
                                   (1 , MapString.empty)
                                   sl
 in let subst = function
     | Dat z -> Dat z
     | Cst x -> Cst x
     | Var y -> Var (try (MapString.find y sigma)
                     with Not_found -> failwith (y ^ "was not found")
                    )
 in mapSuper subst




type 'a proc = Proc of string  *  int * 'a option
type 'a prog = Prog of ('a proc) list * 'a

let proc a b c = Proc (a,b,c)
let prog a b   = Prog (a,b)

let mapProc f (Proc (s,n,x)) = Proc (s,n, omap f x)
let mapProg f (Prog (l ,x) ) = Prog ((List.map (mapProc f) l) , f x)



(*** Recursion ***)

let recY =
 let g = mkFun "x" (mkApp (mkVar "x") (mkVar "x"))
 and h = mkFun "x" (mkApp (mkVar "f") (mkApp (mkVar "x") (mkVar "x")))
 in mkFun "f" (mkApp g h)



let recZ =
    let a = mkFun "y" (multiApp (mkVar "x") [mkVar "x" ; mkVar "y"])
 in let b = mkFun "x" (mkApp (mkVar "f") a)
 in mkFun "f" (mkApp b b)

let addRec c = mkApp (mkFun "rec" c) recZ

(*** Pretty ***)
open Pretty

let pretty_ibs = function
 | Int i      -> string_of_int i
 | Bool true  -> "true"
 | Bool false -> "false"
 | String s   -> "\"" ^ s ^ "\""


let pretty_atom f = function
 | Var s -> pretty_string ("v:" ^ (string_of_int s))
 | Cst s -> pretty_string ("c:" ^ s)
 | Dat z -> pretty_string ("d:" ^ (f z))

let rec pretty_super f = function
 | SAtm a        -> pretty_atom f a
 | SApp (c , cl) -> pretty_origin ( pretty_string "( " >>
                                    pretty_atom f c    >>
                                    pretty_string " "  >>
                                    pretty_origin ( pretty_list (preturn ())
                                                                (pretty_string "\n")
                                                                (List.map (pretty_super f) cl)
                                                  ) >>
                                    pretty_string "\n)"
                                  )
let pretty_proc f (Proc (n,nargs,app)) =
    pretty_origin ( pretty_string (n ^ ":" ^ (string_of_int nargs) ^ " = ") >>
                    (match app with
                      | Some x -> pretty_super f x
                      | None   -> pretty_string "extern"
                    )
                  )


let pretty_prog f (Prog (l,a)) =
 pretty_origin ( pretty_list (preturn ())
                             (pretty_string "\n")
                             (List.map (pretty_proc f) l) >>
                 pretty_string "\nMain = " >>
                 pretty_super f a
               )



(*** Compilation ***)
open Statem


let registerSuper sl c = sget >>= (fun (m,i) ->
                            let name = "sc" ^ (strBaseTwo i)
                         in let m2   = MapString.add name (List.length sl , alphaSuper sl c) m
                         in sput (m2,i + 1) >> sreturn name )



let rec superify u =
 let rec accFun l = function
   | Fun ( s , c ) -> accFun (s :: l) c
   | t             -> let fv = SetString.elements (
                                 List.fold_left (swap SetString.remove) (freeVar t) l
                               )
                      in  superify t                           >>= (fun t' ->
                          registerSuper (fv @ (List.rev l)) t' >>= (fun n  ->
                          sreturn (multiApp (Atm (Cst n))
                                            (List.map mkVar fv)
                                  ) ) )
 and     accApp l = function
   | App (c , d)    -> superify d >>= (fun d' ->
                       accApp (d' :: l) c )
   | Atm a          -> sreturn (SApp (a , l))
   | Fun(_,_) as t  -> accFun [] t >>= (fun t' ->
                       accApp l t' )
 in (match u with
      | Atm a     -> sreturn (SAtm a)
      | Fun (_,_) -> accFun [] u >>= (fun u' -> superify u')
      | App (_,_) -> accApp [] u
    )
                    



let compRun (StateM m) =
 let (t,(m',_)) = m (MapString.empty , 0)
 in Prog (List.map (fun (a,(b,c)) -> Proc (a,b, Some c)) (mapStringBindings m') , alphaSuper [] t)


let progOfAst dorec (AAst (el,t)) =
 let  Prog (scl , sc) = let c  = coreOfTerm t
                        in let c2 = if dorec then addRec c else c
                        in compRun ( superify ( genConstant c2 ) )
 and el2 = List.map (fun (s,i) -> Proc (s,i,None)) (primitives @ el)
 in Prog ( el2 @ scl , sc)
