(*
    Copyright (C) 2012 Christophe Calvès

    This file is part of WC.

    WC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    WC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
*)


(************************
 *                      *
 *       MESSAGES       *
 *                      *
 ************************)

let version =
"WC 0.0.1\n
Licensed under the terms of the GPLv3+\n\
Copyright GTof\n\
\n"


let usagemsg    =
"\
Usage: " ^ (Filename.basename Sys.argv.(0)) ^ " OPTIONS

 -t, --TeX                : compiles to TeX
 -s, --shell              : compiles to POSIX shell
 -S, --super              : compiles to supercombinators
 -i, --intput <file path> : input  file ( or - for stdio )
 -o, --output <file path> : output file ( or - for stdout)
 -v, --version            : prints version informations
 -r, --rec                : enable  recursion support (default)
 -R, --norec              : disable recursion support

"

let texdesc     = "compiles to TeX"
let shelldesc   = "compiles to POSIX shell"
let superdesc   = "compiles to supercombinators"
let versiondesc = "prints version information"
let inputdesc   = "input file path (or - for standard input)"
let outputdesc  = "output file path (or - for standard output)"
let withrecdesc = "enable recursion support (default)"
let norecdesc   = "diable recursion support"



(************************
 *                      *
 *        PARAMS        *
 *                      *
 ************************)


let null _ = ()

type process = TeX
             | Shell
             | Super
             | Version
             | Usage
let process = ref Usage
let setProcess x _ = (process := x)

let inputfile  = ref "-"
let outputfile = ref "-"
let dorec      = ref true

(************************
 *                      *
 *       PROCESSING     *
 *                      *
 ************************)


let upIn f () = 
    let (c,close) = (match !inputfile with
                       | "-" -> (stdin , null)
                       | x   -> let c = open_in x
                                in (c , (fun () -> close_in c))
                    )
 in let ast = Parser.ast Lexer.lexer (Lexing.from_channel c)
 in let _   = close ()
 in f (Comp.progOfAst !dorec ast)

let upOut f x =
    let (c,close) = (match !outputfile with
                       | "-" -> (stdout , null)
                       | x   -> let c = open_out x
                                in (c , (fun () -> close_out c))
                    )
 in (output_string c (f x) ; flush c ; close ())

let processSuper () =
 let f x = Pretty.pretty_run (Comp.pretty_prog Comp.pretty_ibs x)
 in upIn (upOut f) ()

let exedir () =
    let s = Sys.getcwd ()
 in let _ = Sys.chdir (Filename.dirname Sys.argv.(0))
 in let r = Sys.getcwd ()
 in (Sys.chdir s ; r)

let processTeX   () =
 let libpath = (exedir ()) ^ Filename.dir_sep ^ "tex" ^  Filename.dir_sep
 in upOut (upIn (Tex.standAloneOfProg Tex.ofIBS libpath)) ()

let processShell   () =
 let libpath = (exedir ()) ^ Filename.dir_sep ^ "shell" ^  Filename.dir_sep
 in upOut (upIn (Shell.standAloneOfProg Shell.ofIBS libpath)) ()

(************************
 *                      *
 *         MAIN         *
 *                      *
 ************************)


let main () =
 let _ =  Arg.parse [ ( "-t"        , Arg.Unit (setProcess TeX    ) , texdesc    )
                    ; ( "--TeX"     , Arg.Unit (setProcess TeX    ) , texdesc    )
                    ; ( "-s"        , Arg.Unit (setProcess Shell  ) , shelldesc  )
                    ; ( "--shell"   , Arg.Unit (setProcess Shell  ) , shelldesc  )
                    ; ( "-S"        , Arg.Unit (setProcess Super  ) , superdesc  )
                    ; ( "--super"   , Arg.Unit (setProcess Super  ) , superdesc  )
                    ; ( "-v"        , Arg.Unit (setProcess Version) , versiondesc)
                    ; ( "--version" , Arg.Unit (setProcess Version) , versiondesc)
                    ; ( "-i"        , Arg.Set_string inputfile      , inputdesc  )
                    ; ( "--input"   , Arg.Set_string inputfile      , inputdesc  )
                    ; ( "-o"        , Arg.Set_string outputfile     , outputdesc )
                    ; ( "--output"  , Arg.Set_string outputfile     , outputdesc )
                    ; ( "-r"        , Arg.Set        dorec          , withrecdesc)
                    ; ( "--rec"     , Arg.Set        dorec          , withrecdesc)
                    ; ( "-R"        , Arg.Clear      dorec          , norecdesc  )
                    ; ( "--norec"   , Arg.Clear      dorec          , norecdesc  )
                    ]
                    null
                    usagemsg
 in let meth = (match !process with
                  | Super   -> processSuper
                  | TeX     -> processTeX
                  | Shell   -> processShell
                  | Version -> upOut (Base.cst version )
                  | Usage   -> upOut (Base.cst usagemsg)
              )
 in meth ()
;;

main ()



